package zero1hd.rhythmbullet.desktop.screens;

import com.badlogic.gdx.Gdx;
import com.badlogic.gdx.InputMultiplexer;
import com.badlogic.gdx.Screen;
import com.badlogic.gdx.ScreenAdapter;
import com.badlogic.gdx.graphics.GL20;
import com.badlogic.gdx.graphics.Pixmap.Format;
import com.badlogic.gdx.graphics.g2d.Batch;
import com.badlogic.gdx.graphics.g2d.SpriteBatch;
import com.badlogic.gdx.graphics.g2d.TextureRegion;
import com.badlogic.gdx.graphics.glutils.FrameBuffer;
import com.badlogic.gdx.graphics.glutils.ShaderProgram;
import com.badlogic.gdx.scenes.scene2d.Actor;
import com.badlogic.gdx.scenes.scene2d.ui.ImageButton;
import com.badlogic.gdx.scenes.scene2d.utils.ChangeListener;
import com.badlogic.gdx.utils.viewport.ScreenViewport;

import zero1hd.rhythmbullet.RhythmBullet;
import zero1hd.rhythmbullet.desktop.audio.map.GamePlayMap;
import zero1hd.rhythmbullet.desktop.stages.GameHUD;
import zero1hd.rhythmbullet.desktop.stages.GamePlayArea;
import zero1hd.rhythmbullet.util.MusicManager;

public class GameScreen extends ScreenAdapter {
	
	protected GamePlayArea gameArea;
	public GameHUD gameHUD;
	
	protected InputMultiplexer inputs;
	public RhythmBullet core;
	
	private MusicManager music;
	
	private ShaderProgram gaussianBlurShader;
	private ShaderProgram brightFilterShader;
	private ShaderProgram combineShader;
	private FrameBuffer lightFilterBuffer;
	private FrameBuffer normalBuffer;
	private FrameBuffer hBlur, vBlur;
	private TextureRegion fboRegion;
	private int fboSize;
	private Batch batch;
	private ScreenViewport screenViewport;
	private int blurlvl = 4;
	
	/**
	 * The game screen where the game play area, and hud are placed.
	 * @param core The game context object
	 * @param screen the screen to return to if player decides to press "quit" button.
	 */
	public GameScreen(RhythmBullet core, Screen screen) {
		this.core = core;
		
		// Overlay stuff
		ImageButton pause = new ImageButton(core.getDefaultSkin().getDrawable("pause"),
				core.getDefaultSkin().getDrawable("pause-down"));
		pause.setPosition(Gdx.graphics.getWidth() - pause.getWidth() - 25,
				Gdx.graphics.getHeight() - pause.getHeight() - 25);
		pause.addListener(new ChangeListener() {
			@Override
			public void changed(ChangeEvent event, Actor actor) {
				pause();
			}
		});
		
		gameArea = new GamePlayArea(core.getAssetManager(), core.getPrefs());
		gameHUD = new GameHUD(core.getDefaultSkin(), 100f, gameArea, screen, core);
		inputs = new InputMultiplexer();
		inputs.addProcessor(gameHUD);
		inputs.addProcessor(gameArea);
		
	}
	
	public void setGamePlayMap(GamePlayMap gpm) {
		music = gpm.getMusicData();
		gameArea.setAudioMap(gpm);
		gameHUD.setMusic(gpm.getMusicData());
	}
	
	@Override
	public void show() {
		Gdx.input.setInputProcessor(inputs);
		loadShaders();
		super.show();
	}
	
	/**
	* needs to be called right after set as screen (should be called in show method).
	* This is due to the saving preference being done once the screen is hidden.
	* @param prefs
	*/
	public void loadShaders() {
		if (core.getPrefs().getBoolean("glow shader")) {
			batch = new SpriteBatch();
			screenViewport = new ScreenViewport();
			
			Gdx.app.debug("Shader", "using glow shader");
			brightFilterShader = new ShaderProgram(Gdx.files.internal("shaders/basic.vsh"), Gdx.files.internal("shaders/bright_filter.fsh"));
			if (!brightFilterShader.isCompiled()) {
				Gdx.app.error("Shader failed to compile", brightFilterShader.getLog());
				System.exit(0);
			}
			if (brightFilterShader.getLog().length() != 0) {
				Gdx.app.error("Shader", brightFilterShader.getLog());
			}
				
			gaussianBlurShader = new ShaderProgram(Gdx.files.internal("shaders/basic.vsh"), Gdx.files.internal("shaders/gaussian_blur.fsh"));
			if (!gaussianBlurShader.isCompiled()) {
				Gdx.app.error("Shader failed to compile", gaussianBlurShader.getLog());
				System.exit(0);
			}
			if (gaussianBlurShader.getLog().length() != 0) {
				Gdx.app.error("Shader", gaussianBlurShader.getLog());
			}
			
			combineShader = new ShaderProgram(Gdx.files.internal("shaders/basic.vsh"), Gdx.files.internal("shaders/combine.fsh"));
			if (!combineShader.isCompiled()) {
				Gdx.app.error("Shader failed to compile", combineShader.getLog());
				System.exit(0);
			}
			if (combineShader.getLog().length() != 0) {
				Gdx.app.error("Shader", combineShader.getLog());
			}
			
			
			if (Gdx.graphics.getWidth() < 1024) {
				fboSize = 2048;
			} else if (Gdx.graphics.getWidth() < 2048) {
				fboSize = 4096;
			} else {
				fboSize = 4096;
			}
			
			lightFilterBuffer = new FrameBuffer(Format.RGBA8888, fboSize/2, fboSize/2, false);
			normalBuffer = new FrameBuffer(Format.RGBA8888, fboSize, fboSize, false);
			hBlur = new FrameBuffer(Format.RGBA8888, fboSize/2, fboSize/2, false);
			vBlur = new FrameBuffer(Format.RGBA8888, fboSize/2, fboSize/2, false);

			fboRegion = new TextureRegion(normalBuffer.getColorBufferTexture());
			fboRegion.flip(false, true);

			combineShader.begin();
			combineShader.setUniformi("u_texture1", 1);
			combineShader.end();
			
			vBlur.getColorBufferTexture().bind(1);
			
			Gdx.gl.glActiveTexture(GL20.GL_TEXTURE0);
		}
		
		ShaderProgram.pedantic = false;
	}
	
	@Override
	public void render(float delta) {
		Gdx.gl.glClearColor(0f, 0f, 0f, 1f);
		Gdx.gl.glClear(GL20.GL_COLOR_BUFFER_BIT);
		//actual game and hud
		if (!gameHUD.isPaused()) {
			gameArea.act(delta);
			if (gameArea.getPolyjetEntity().isDead()) {
				end(false);
			}
			gameHUD.act(delta);
		}
		gameArea.getViewport().apply();
		
		blurlvl = 4;
		if (gaussianBlurShader != null) {
			//Begin drawing a normal version of screen
			gameArea.getViewport().apply();
			normalBuffer.begin();
			Gdx.gl.glClearColor(0.22f, 0f, 0f, 1f);
			Gdx.gl.glClear(GL20.GL_COLOR_BUFFER_BIT);
			gameArea.draw();
			normalBuffer.end();
			//END STAGE BATCH
			
			//BEGINNING NORMAL SCREEN RENDER
			screenViewport.apply();
			//Begin light filtering
			lightFilterBuffer.begin();
			Gdx.gl.glClearColor(0f, 0f, 0f, 1f);
			Gdx.gl.glClear(GL20.GL_COLOR_BUFFER_BIT);
			fboRegion.setTexture(normalBuffer.getColorBufferTexture());
			batch.setShader(brightFilterShader);
			batch.begin(); //BATCH STARTS HERE
			batch.draw(fboRegion, 0, 0, gameArea.getWidth(), gameArea.getHeight());
			batch.flush();
			lightFilterBuffer.end();
			
			for (int i = 0; i < blurlvl; i++) {
				//Horizontal gaussian blur
				hBlur.begin();
				if (i > 0) {
					fboRegion.setTexture(vBlur.getColorBufferTexture());
				} else {
					fboRegion.setTexture(lightFilterBuffer.getColorBufferTexture());
				}				batch.setShader(gaussianBlurShader);
				gaussianBlurShader.setUniformi("horizontal", 1);
				batch.draw(fboRegion, 0f, 0f, gameArea.getWidth(), gameArea.getHeight());
				batch.flush();
				hBlur.end();
				
				//Vertical gaussian blur
				vBlur.begin();
				fboRegion.setTexture(hBlur.getColorBufferTexture());
				batch.setShader(gaussianBlurShader);
				gaussianBlurShader.setUniformi("horizontal", 0);
				batch.draw(fboRegion, 0f, 0f, gameArea.getWidth(), gameArea.getHeight());
				batch.flush();
				vBlur.end();
			}

			//draw a final copy to a fbo
			Gdx.gl.glClearColor(0f, 0f, 0f, 0f);
			Gdx.gl.glClear(GL20.GL_COLOR_BUFFER_BIT);
			batch.setShader(combineShader);
			fboRegion.setTexture(normalBuffer.getColorBufferTexture());
			batch.draw(fboRegion, 0f, 0f, gameArea.getWidth(), gameArea.getHeight());
			batch.setShader(null);
			batch.end(); //STAGE BATCH ENDS HERE

		} else {
			gameArea.draw();
		}
		
		gameHUD.getViewport().apply();
		gameHUD.draw();
		if (music != null && !music.isPlaying()) {
			end(true);
		}
		
		super.render(delta);
	}

	private void end(boolean trueEnd) {
		if (trueEnd) {
			// TODO they didn't die
		} else {
			// TODO They done goofed and ended up dying.
		}
	}

	@Override
	public void dispose() {
		gameHUD.dispose();
		gameArea.dispose();
		
		if (gaussianBlurShader != null) {
			normalBuffer.dispose();
			lightFilterBuffer.dispose();
			hBlur.dispose();
			vBlur.dispose();
			
			brightFilterShader.dispose();
			gaussianBlurShader.dispose();
			combineShader.dispose();
			gaussianBlurShader.dispose();
		}
		super.dispose();
	}
	
	@Override
	public void pause() {
		gameHUD.setPaused(true);
		super.pause();
	}
	
	@Override
	public void resize(int width, int height) {
		gameArea.getViewport().update(width, height, true);
		gameHUD.getViewport().update(width, height, true);

		super.resize(width, height);
	}
	
	public GamePlayArea getGameArea() {
		return gameArea;
	}
	
	public void play() {
		if (music == null) throw new NullPointerException("Idiot, you can't have a music game not have music on the gameplay screen...");
		Gdx.input.setInputProcessor(inputs);
		if (!gameHUD.isPaused()) {
			music.play();
		}
	}
}